from __future__ import annotations

import datetime
from typing import Any, Callable, Generator, Union

import pendulum
from pendulum.tz.timezone import Timezone
from pydantic.datetime_parse import parse_datetime


class DateTime(pendulum.DateTime):
    @classmethod
    def __get_validators__(cls) -> Generator[Callable[[Any], DateTime], None, None]:
        yield parse_datetime
        yield cls._to_dt

    @classmethod
    def _to_dt(cls, v: Union[datetime.datetime, pendulum.DateTime]) -> DateTime:
        if isinstance(v, datetime.datetime):
            return cls.from_datetime(v)
        return v

    @classmethod
    def from_datetime(cls, dt: datetime.datetime, tz: Timezone = pendulum.UTC) -> DateTime:
        return pendulum.instance(dt=dt, tz=tz)



from __future__ import annotations

from typing import Any, ClassVar, Generic, Optional, TypeVar, get_args

import pymongo
from pymongo import UpdateOne
from pymongo.collection import Collection
from pymongo.database import Database

from .. import types
from ..entity import Entity
from ..pagination import Pagination
from ..query.mongo import MongoFilterQuery
from . import base

EntityType = TypeVar("EntityType", bound=Entity)


class MongoRepository(base.BaseRepository, Generic[EntityType]):
    __collection_name__: ClassVar[str] = ""
    __indexes__: ClassVar[Optional[list[pymongo.IndexModel]]] = None

    def __init__(self, collection: Collection):
        self.collection = collection

    @classmethod
    def create(cls, database: Database, **kwargs) -> MongoRepository:
        collection = database.get_collection(cls.__collection_name__, **kwargs)
        if cls.__indexes__ is not None:
            collection.create_indexes(cls.__indexes__)
        return cls(collection)

    def find_by_id(self, entity_id: types.PyObjectId, **kwargs) -> Optional[EntityType]:
        return self.find_by({"_id": entity_id}, **kwargs)

    def find_by(self, filter_: dict[str, Any], **kwargs) -> Optional[EntityType]:
        """https://stackoverflow.com/a/73746554/9331155"""
        entity_type: EntityType = get_args(self.__class__.__orig_bases__[0])[0]
        result = self.collection.find_one(filter=filter_, **kwargs)
        return result and entity_type.parse_obj(result)

    def add(self, entity: EntityType, **kwargs) -> None:
        self.collection.insert_one(entity.dict(by_alias=True), **kwargs)

    def add_many(self, entities: list[EntityType], **kwargs) -> None:
        self.collection.insert_many([entity.dict(by_alias=True) for entity in entities], **kwargs)

    def update(self, entity: EntityType, **kwargs) -> None:
        self.collection.update_one(
            {"_id": entity.id},
            {"$set": entity.dict(by_alias=True)},
            **kwargs,
        )

    def update_many(self, entities: list[EntityType], **kwargs) -> None:
        self.collection.bulk_write(
            requests=[
                UpdateOne({"_id": entity.id}, {"$set": entity.dict(by_alias=True)})
                for entity in entities
            ],
            **kwargs,
        )

    def delete(self, entity_id: types.PyObjectId, **kwargs) -> None:
        self.collection.delete_one({"_id": entity_id}, **kwargs)

    def execute_raw(self, operation: str, **operation_kwargs) -> Any:
        op = getattr(self.collection, operation, None)
        if op is None:
            raise ValueError(f"Unknown operation on collection: {operation}")
        return op(**operation_kwargs)

    def filter(self, qry: MongoFilterQuery) -> Pagination:
        result = self.collection.aggregate(qry.to_query())
        result = next(result, None)
        return Pagination(
            total=(result["metadata"][0]["total"] if result["metadata"] else 0),
            items=result["items"],
            page=qry.page,
            per_page=qry.per_page,
        )



from __future__ import annotations

from typing import Any, Generic, TypeVar, get_args

from pydantic import ValidationError
from pydantic.generics import GenericModel
from pydantic.typing import get_origin, is_union

from .. import container
from ..pagination import Pagination
from .extra import schema_by_field_name

ListItem = TypeVar("ListItem")


class Schema(container.BaseModel):
    class Config:
        schema_extra = schema_by_field_name()


class TimeStampedSchema(container.TimeStampedModel, Schema):
    ...


class ListSchema(Schema, GenericModel, Generic[ListItem]):
    total: int
    items: list[ListItem]

    @classmethod
    def from_pagination(
        cls,
        pagination: Pagination,
        parser_name: str = "parse_obj",
        **parser_kwargs: Any,
    ) -> ListSchema:
        if not pagination.items:
            return cls(**pagination.dict())

        item_type = cls.__fields__["items"].type_
        item_origin = get_origin(item_type)
        if is_union(item_origin):
            for arg in get_args(item_type):
                try:
                    arg.parse_obj(pagination.items[0])
                    item_type = arg
                except ValidationError:
                    continue

        item_parsable = issubclass(item_type, container.BaseModel)
        if not item_parsable:
            data = pagination.dict(exclude={"extra"}) | pagination.extra
            return cls(**data)

        if (parser := getattr(item_type, parser_name, None)) is None:
            raise AttributeError(f"{item_type.__name__} has no attribute: {parser_name}")

        return cls(
            **pagination.dict(exclude={"items", "extra"}),
            **pagination.extra,
            items=[parser(item, **parser_kwargs) for item in pagination.items],
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['compose',
 'compose.query',
 'compose.query.mongo',
 'compose.repository',
 'compose.schema',
 'compose.types']

package_data = \
{'': ['*']}

install_requires = \
['pendulum>=2.1.2,<3.0.0',
 'pydantic>=1.10.2,<2.0.0',
 'pymongo[aws]>=4.3.3,<5.0.0']

setup_kwargs = {
    'name': 'pozalabs-compose',
    'version': '0.2.0',
    'description': 'Backend components for POZAlabs',
    'long_description': 'None',
    'author': 'pozalabs',
    'author_email': 'contact@pozalabs.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



from typing import Any, Optional

from pydantic import conint

from .. import base


class MongoQuery(base.Query):
    def to_query(self) -> list[dict[str, Any]]:
        ...


class MongoFilterQuery(MongoQuery):
    page: Optional[conint(ge=1)] = None
    per_page: Optional[conint(ge=1)] = None

    @property
    def can_paginate(self) -> bool:
        return self.page is not None and self.per_page is not None



from __future__ import annotations

from typing import Any, Optional, Union

import bson
from pydantic import BaseModel as PydanticBaseModel

from . import field, types


class BaseModel(PydanticBaseModel):
    def copy(
        self,
        *,
        include: Optional[Union[set[str], dict[str, Any]]] = None,
        exclude: Optional[Union[set[str], dict[str, Any]]] = None,
        update: Optional[dict[str, Any]] = None,
        deep: bool = False,
        strict: bool = True,
    ) -> BaseModel:
        if strict and (diff := set((update or {}).keys()) - self.__fields_set__):
            raise AttributeError(f"{self.__class__.__name__} has no attributes: {', '.join(diff)}")

        return super().copy(
            include=include,
            exclude=exclude,
            update=update,
            deep=deep,
        )

    class Config:
        json_encoders = {bson.ObjectId: str}
        allow_population_by_field_name = True
        validate_assignment = True


class TimeStampedModel(BaseModel):
    created_at: types.DateTime = field.DateTimeField()
    updated_at: types.DateTime = field.DateTimeField()

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)

        created_at_field = cls.__fields__.pop("created_at")
        updated_at_field = cls.__fields__.pop("updated_at")
        cls.__fields__ |= dict(created_at=created_at_field, updated_at=updated_at_field)
